"use strict";
this.haketilo_exports = this.haketilo_exports || {};
this.haketilo_exports["background/webrequest.js"] = {};

(function() {
var globalThis = this.haketilo_this
{
/**
 * This file is part of Haketilo.
 *
 * Function: Modify HTTP traffic usng webRequest API.
 *
 * Copyright (C) 2021, 2022 Wojtek Kosior <koszko@koszko.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * As additional permission under GNU GPL version 3 section 7, you
 * may distribute forms of that code without the copy of the GNU
 * GPL normally required by section 4, provided you include this
 * license notice and, in case of non-source distribution, a URL
 * through which recipients can access the Corresponding Source.
 * If you modify file(s) with this exception, you may extend this
 * exception to your version of the file(s), but you are not
 * obligated to do so. If you do not wish to do so, delete this
 * exception statement from your version.
 *
 * As a special exception to the GPL, any HTML file which merely
 * makes function calls to this code, and for that purpose
 * includes it by reference shall be deemed a separate work for
 * copyright law purposes. If you modify this code, you may extend
 * this exception to your version of the code, but you are not
 * obligated to do so. If you do not wish to do so, delete this
 * exception statement from your version.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *
 * I, Wojtek Kosior, thereby promise not to sue for violation of this file's
 * license. Although I request that you do not make use of this code in a
 * proprietary program, I am not going to enforce this in court.
 */

const haketilodb = haketilo_exports["common/indexeddb.js"];

const stream_filter = haketilo_exports["background/stream_filter.js"];

const browser = haketilo_exports["common/browser.js"].browser;
const is_privileged_url = haketilo_exports["common/misc.js"].is_privileged_url;
const csp_header_regex = haketilo_exports["common/misc.js"].csp_header_regex;
const sha256 = haketilo_exports["common/misc.js"].sha256_async;
const decide_policy = haketilo_exports["common/policy.js"].decide_policy;

const tree = haketilo_exports["background/patterns_query_manager.js"].tree;
const default_allow = haketilo_exports["background/patterns_query_manager.js"].default_allow;

let secret;

/*
 * Under Mozilla-based browsers, responses are cached together with headers as
 * they appear *after* modifications by Haketilo. This means Haketilo's CSP
 * script-blocking headers might be present in responses loaded from cache. In
 * the meantime the user might have changes Haketilo settings to instead allow
 * the scripts on the page in question. This causes a problem and creates the
 * need to somehow restore the response headers to the state in which they
 * arrived from the server.
 * To cope with this, Haketilo will inject some additional headers with private
 * data. Those will include a hard-to-guess value derived from extension's
 * internal ID. It is assumed the internal ID has a longer lifetime than cached
 * responses.
 */

const settings_page_url = browser.runtime.getURL("html/settings.html");
const header_prefix_prom = sha256(settings_page_url)
      .then(hash => `X-Haketilo-${hash}`);

/*
 * Mozilla, unlike Chrome, allows webRequest callbacks to return promises. Here
 * we leverage that to be able to use asynchronous sha256 computation.
 */
async function on_headers_received(details) {
    const url = details.url;
    if (is_privileged_url(details.url))
	return;

    let headers = details.responseHeaders;

    const prefix = await header_prefix_prom;

    /*
     * We assume that the original CSP headers of a response are always
     * preserved under names of the form:
     *     X-Haketilo-<some_secret>-<original_name>
     * In some cases the original response may contain no CSP headers. To still
     * be able to tell whether the headers we were provided were modified by
     * Haketilo in the past, all modifications are accompanied by addition of an
     * extra header with name:
     *     X-Haketilo-<some_secret>
     */

    const restore_old_headers = details.fromCache &&
	  !!headers.filter(h => h.name === prefix).length;

    if (restore_old_headers) {
	const restored_headers = [];

	for (const h of headers) {
	    if (csp_header_regex.test(h.name) || h.name === prefix)
		continue;

	    if (h.name.startsWith(prefix)) {
		restored_headers.push({
		    name:  h.name.substring(prefix.length + 1),
		    value: h.value
		});
	    } else {
		restored_headers.push(h);
	    }
	}

	headers = restored_headers;
    }

    const policy =
	  decide_policy(tree, details.url, !!default_allow.value, secret);

    if (!policy.allow) {
	const to_append = [{name: prefix, value: ":)"}];

	for (const h of headers.filter(h => csp_header_regex.test(h.name))) {
	    if (!policy.payload)
		to_append.push(Object.assign({}, h));

	    h.name = `${prefix}-${h.name}`;
	}

	headers.push(...to_append);

	headers.push({name: "Content-Security-Policy", value: policy.csp});
    }

    /*
     * When page is meant to be viewed in the browser, use streamFilter to
     * inject a dummy <script> at the very beginning of it. This <script>
     * will cause extension's content scripts to run before page's first <meta>
     * tag is rendered so that they can prevent CSP rules inside <meta> tags
     * from blocking the payload we want to inject.
     */

    let use_stream_filter = !!policy.payload;
    if (use_stream_filter) {
	for (const header of headers) {
	    if (header.name.toLowerCase().trim() !== "content-disposition")
		continue;

	    if (/^\s*attachment\s*(;.*)$/i.test(header.value)) {
		use_stream_filter = false;
	    } else {
		use_stream_filter = true;
		break;
	    }
	}
    }
    use_stream_filter = use_stream_filter &&
	(details.statusCode < 300 || details.statusCode >= 400);

    if (use_stream_filter)
	headers = stream_filter.apply(details, headers, policy);

    return {responseHeaders: headers};
}


async function start(secret_)
{
    secret = secret_;

    const extra_opts = ["blocking"];

    browser.webRequest.onHeadersReceived.addListener(
	on_headers_received,
	{urls: ["<all_urls>"], types: ["main_frame", "sub_frame"]},
	extra_opts.concat("responseHeaders")
    );

}
this.haketilo_exports["background/webrequest.js"].start = (start);
}
}).call({
    haketilo_exports: this.haketilo_exports,
    haketilo_this:    this
});
